﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: IniFile.cs 557 2012-09-12 11:38:47Z cr333 $

Based on:
  - <http://www.codeproject.com/Messages/3083693/Code-Update.aspx>
  - <http://www.pinvoke.net/default.aspx/kernel32.getprivateprofilestring>
  - <http://www.codeproject.com/KB/files/INI_File_Enumerator.aspx>
*/
using System.Globalization;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;

namespace RGBZcamera.Library.Config
{
	/// <summary>Reads data in from or writes data to INI files.</summary>
	public class IniFile
	{
		#region Imports

		[DllImport("kernel32.dll", EntryPoint = "GetPrivateProfileStringW", SetLastError = true, CharSet = CharSet.Unicode, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
		private static extern int GetPrivateProfileString(string section, string key, string value, StringBuilder result, int size, string filename);

		[DllImport("kernel32.dll", EntryPoint = "GetPrivateProfileStringW", SetLastError = true, CharSet = CharSet.Unicode, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
		private static extern int GetPrivateProfileString(string section, string key, string value, string result, int size, string filename);

		[DllImport("kernel32.dll", EntryPoint = "WritePrivateProfileStringW", SetLastError = true, CharSet = CharSet.Unicode, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
		private static extern int WritePrivateProfileString(string section, string key, string value, string filename);

		#endregion

		/// <summary>Gets or sets the path to the INI file.</summary>
		public string Path { get; set; }

		public CultureInfo Culture { get; set; }

		/// <summary>Initialises a new instance of the <see cref="IniFile"/> class.</summary>
		/// <param name="path">The path.</param>
		public IniFile(string path)
		{
			if (!File.Exists(path))
				throw new FileNotFoundException("The INI file was not found.", path);

			Path = path;
			Culture = CultureInfo.CreateSpecificCulture("en-GB");
		}

		#region Core interface

		/// <summary>Writes a key-value pair in the given section. [untested]</summary>
		/// <param name="section">The section name.</param>
		/// <param name="key">The key name.</param>
		/// <param name="value">The value string.</param>
		public void WriteString(string section, string key, string value)
		{
			WritePrivateProfileString(section, key, value, Path);
		}

		/// <summary>Reads the value of a key from the given section.</summary>
		/// <param name="section">The section name.</param>
		/// <param name="key">The key name.</param>
		/// <returns>The read value.</returns>
		public string ReadString(string section, string key)
		{
			StringBuilder sb = new StringBuilder(256);
			GetPrivateProfileString(section, key, "", sb, sb.Capacity, Path);
			string value = sb.ToString();

			// strip # comments at end of lines
			var hashIndex = value.IndexOf('#');
			if (hashIndex != -1)
				value = value.Remove(hashIndex);

			return value;
		}

		/// <summary>Returns the names of all sections found in the INI file.</summary>
		/// <returns>Section names.</returns>
		public string[] GetSections()
		{
			return GetKeys(null);
		}

		/// <summary>Returns the names of all keys found in the given section.</summary>
		/// <param name="section">The section name. If null, returns all section names.</param>
		/// <returns>Array of keys.</returns>
		public string[] GetKeys(string section)
		{
			// start with 512 byte buffer; keep doubling size if required
			for (int maxsize = 512; true; maxsize *= 2)
			{
				string bytes = new string(' ', maxsize);
				int size = GetPrivateProfileString(section, null, null, bytes, maxsize, Path);

				// if the strings don't fit into the buffer, retry with doubled buffer size
				if (size >= maxsize - 2) continue;

				// cut off string at the end
				string Selected = bytes.Substring(0, size - (size > 0 ? 1 : 0));

				// split string at zeros
				return Selected.Split('\0');
			}
		}

		#endregion Core interface

		#region Extended interface

		/// <summary>Reads a 32-bit integer from the given section and key. 0 in case of error.</summary>
		/// <param name="section">The section name.</param>
		/// <param name="key">The key name.</param>
		/// <returns>The read value.</returns>
		public int ReadInt(string section, string key)
		{
			int value;
			if (int.TryParse(ReadString(section, key), NumberStyles.Number, Culture, out value))
				return value;
			else
				return 0;
		}

		/// <summary>Reads a floating-point number from the given section and key. 0.0f in case of error.</summary>
		/// <param name="section">The section name.</param>
		/// <param name="key">The key name.</param>
		/// <returns>The read value.</returns>
		public float ReadFloat(string section, string key)
		{
			float value;
			if (float.TryParse(ReadString(section, key), NumberStyles.Float, Culture, out value))
				return value;
			else
				return 0.0f;
		}

		/// <summary>Reads a tuple of comma-separated floating-point numbers from the given section and key.</summary>
		/// <param name="section">The section name.</param>
		/// <param name="key">The key name.</param>
		/// <returns>The read values.</returns>
		public float[] ReadFloatTuple(string section, string key)
		{
			// read in config string and check for empty string
			var contents = ReadString(section, key);
			if(contents == "") return new float[0];

			// split config string into string[]
			var strings = contents.Split(',', ';');

			// trim spaces, parse as float and convert to float[]
			return strings.Select(x => float.Parse(x.Trim(' ', '\t', 'f', 'F'), Culture)).ToArray();
		}

		#endregion Extended interface
	}
}
