﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: FillInJointBilateralFilter.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;
using VideoLib.OpenGL;
using OpenTK;

namespace RGBZcamera.Library.FillIn
{
	/// <summary>Applies a joint-bilateral filter for filling in invalid pixels (w=0) while preserving edges in the colour image.</summary>
	public class FillInJointBilateralFilter : FboNode
	{
		/// <summary>Gets or sets the spatial sigma.</summary>
		[Uniform(), Description("Standard deviation for spatial filtering.")]
		public float SpatialSigma { get; set; }

		/// <summary>Gets or sets the range sigma.</summary>
		[Uniform(), Description("Standard deviation for range filtering.")]
		public float RangeSigma { get; set; }

		/// <summary>Gets or sets whether points filled in lower levels are recomputed in higher levels.</summary>
		[Uniform(), Description("Recompute points filled in lower levels.")]
		public bool RecomputePoints { get; set; }

		[Uniform()]
		public bool OnOff { get; set; }

		[Uniform()]
		public bool Debug { get; set; }

		[Uniform()]
		public bool OverlayColour { get; set; }

		[Uniform()]
		public bool OverlayWeights { get; set; }

		[Uniform("weightScale")]
		public float OverlayWeightsScale { get; set; }

		[Uniform("weightPoint")]
		public Vector2 OverlayWeightsPoint { get; set; }

		[Uniform("threshMap"),DisplayName("Threshold Z")]
		public bool UseThresholdDistanceMapAsImage { get; set; }

		/// <summary>Initialises a new instance of the <see cref="FillInJointBilateralFilter"/> class.</summary>
		public FillInJointBilateralFilter()
		{
			// default settings
			SpatialSigma = 4;
			RangeSigma   = 0.1f;
			RecomputePoints = true;
			OnOff = true;
			Debug = false;
			OverlayColour = false;

			OverlayWeights = false;
			OverlayWeightsScale = 1;
			UseThresholdDistanceMapAsImage = false;

			AddInputPins("colour", "points", "rawpoints");
			AddOutputPin("out");
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("FillIn.Shaders.FillInJointBilateralFilter.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Name = "colour",    Unit = 0 },
				new TextureBinding { Name = "points",    Unit = 1 },
				new TextureBinding { Name = "rawpoints", Unit = 2 },
			};

			base.Init();
		}

		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			AssertFrameProperties(inputs, FrameProps.Size);

			// allocate a texture that is identical to the input
			var resultTexture = TextureFrame.CreateToMatch(inputs["points"]);
			resultTexture.Alloc();

			Draw(inputs, resultTexture.Texture);

			return resultTexture;
		}
	}
}
