﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: Relighter.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System.Collections.Generic;
using System.ComponentModel;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib.OpenGL;

namespace RGBZcamera.Library
{
	/// <summary>Environment maps avalable for relighting.</summary>
	public enum EnvironmentMap
	{
		None      = 0,
		Basilica  = 1,
		Beach     = 2,
		Building  = 3,
		Campus    = 4,
		Cathedral = 5,
		Galileo   = 6,
		Grove     = 7,
		Kitchen   = 8,
		Uffizi    = 9,
		PointLight=10,
	}

	/// <summary>Relights a scene using spherical harmonics lighting approximation.</summary>
	public class Relighter : FboNode
	{
		[Uniform("lighting"), DisplayName("Lighting coefficient"), Description("Estimated scene lighting."), ReadOnly(true)]
		public float[] LightingCoefficients { get; set; }

		[Uniform("envMap"), DefaultValue(EnvironmentMap.Grove), DisplayName("Environment map"), Description("Selects the environment map to use for relighting.")]
		public EnvironmentMap EnvironmentMap { get; set; }

		[Uniform, DefaultValue(1.0f), DisplayName("Scale albedo"), Description("The albedo is scaled by this factor, to adapt the apparent light level.")]
		public float ScaleAlbedo { get; set; }

		[Uniform, DisplayName("Rotatation"), Description("Rotate the environment using Euler angles (degrees; y/up, x/left, z/to viewer).")]
		public Vector3 Rotation { get; set; }

		[Uniform, DefaultValue(false), Description("Outputs a matte for compositing in a new background."), DisplayName("Output matte?")]
		public bool OutputMatte { get; set; }

		[Uniform, Description("Background plane equation for background subtraction (Xx+Yy+Zy+W=0)."), DisplayName("Background plane"), Category("Background")]
		public Vector4 BackgroundPlane { get; set; }

		[Uniform, DefaultValue(float.NegativeInfinity), Description("Threshold for background segmentation."), DisplayName("Plane threshold"), Category("Background")]
		public float PlaneThreshold { get; set; }

		public bool ShouldSerializeRotation() { return !Rotation.Equals(new Vector3()); }
		public bool ShouldSerializeBackgroundPlane() { return !BackgroundPlane.Equals(new Vector4(0, 0, -1, 1)); }
		public void ResetRotation() { Rotation = new Vector3(); }
		public void ResetBackgroundPlane() { BackgroundPlane = new Vector4(0, 0, -1, 1); }

		/// <summary>Initialises a new instance of the <see cref="Relighter"/> class.</summary>
		public Relighter()
		{
			LightingCoefficients = new float[] { 1, 0, 0, 0, 0, 0, 0, 0, 0 }; // default: constant ambient
			EnvironmentMap       = EnvironmentMap.Grove;
			ScaleAlbedo          = 1.0f;
			OutputMatte          = false;
			BackgroundPlane      = new Vector4(0, 0, -1, 1); // default: plane at 1m depth from camera
			PlaneThreshold       = float.NegativeInfinity; // disable thresholding

			AddInputPins("colour", "points");
			AddOutputPin("out");
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.Mesh.vert", ShaderType.FragmentShader),
				Utils.CreateShaderFromResource("Shaders.Relighter.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Name = "colour", Unit = 0 },
				new TextureBinding { Name = "points", Unit = 1 },
			};

			base.Init();
		}

		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			AssertFrameProperties(inputs, FrameProps.Size);

			// allocate a texture that is identical to the input
			var resultTexture = TextureFrame.CreateToMatch(inputs["colour"]);
			resultTexture.Alloc();

			// rotate lighting over time
			//RotateEnvironment = 5 * inputs["colour"].SequenceNumber;

			Draw(inputs, resultTexture.Texture);

			return resultTexture;
		}
	}
}
