/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: WinnemoellerQuantisation.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#version 120
#extension GL_EXT_gpu_shader4 : enable

//---- Inputs -------------------------------------------------------------------------------------

// input textures
uniform sampler2D src;

// other parameters
uniform int q; // number of bins
uniform float lambdaDelta;
uniform float omegaDelta;
uniform float lambdaPhi;
uniform float omegaPhi;

//---- Functions ----------------------------------------------------------------------------------

// TANH Function (Hyperbolic Tangent)
// NB: This function is first included in GLSL 1.3 (OpenGL 3.0).
float tanh(float val)
{
	float tmp = exp(val);
	float tanH = (tmp - 1.0 / tmp) / (tmp + 1.0 / tmp);
	return tanH;
}

vec3 tanh(vec3 val)
{
	vec3 tmp = exp(val);
	vec3 tanH = (tmp - 1.0 / tmp) / (tmp + 1.0 / tmp);
	return tanH;
}

//---- Main ---------------------------------------------------------------------------------------
	
void main()
{	
	// find out size of texture/viewport to address individual pixels
	vec2 viewport = textureSize2D(src, 0);
	
	// look up colours (centre, top, right, bottom, left)
	vec3 colour  = texture2D(src, gl_FragCoord.xy / viewport).rgb;
	vec3 colourR = texture2D(src, (gl_FragCoord.xy + vec2( 1.0f,  0.0f)) / viewport).rgb;
	vec3 colourB = texture2D(src, (gl_FragCoord.xy + vec2( 0.0f,  1.0f)) / viewport).rgb;
	
	// compute lightness gradient
	float dX = (colourR.x - colour.x) / 1.0f;
	float dY = (colourB.x - colour.x) / 1.0f;
	//float grad = abs(dX) + abs(dY);
	float grad = sqrt(dX * dX + dY * dY);
	
	// linearly scale gradient from [lambdaDelta, omegaDelta] to [0, 1] (with clamping)
	grad = clamp(grad, lambdaDelta, omegaDelta);
	grad = (grad - lambdaDelta) / (omegaDelta - lambdaDelta);
	
	// linearly scale gradient from [0, 1] to [lambdaPhi, omegaPhi]
	float phiQ = lambdaPhi + grad * (omegaPhi - lambdaPhi);
	
	// nearest bin boundary
	vec3 q_nearest = 100.0 * ivec3(q * colour / 100.0 + 0.5f) / float(q);

	// soft quantisation
	float deltaQ = 100.0 / q;
	vec3 Q = q_nearest + deltaQ / 2.0 * tanh(clamp(phiQ * (colour - q_nearest), -50, 50));

	//gl_FragColor = vec4(vec3(grad, 0, 0), 1); // gradient (debug)
	//gl_FragColor = vec4(Q, 1); // LAB quantisation
	gl_FragColor = vec4(Q.x, colour.yz, 1); // lightness quantisation
}
