﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: XToon.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#version 120
#extension GL_EXT_gpu_shader4 : enable

// Implements X-Toon shader per Barla et al. 2006.

// input textures
uniform sampler2D points;
uniform sampler2D toonTexture;

// which attribute map to use
uniform int attributeMap;

// depth range for scaling depth-based attribute map
uniform vec2 depthRange;

// direction or position of light
uniform vec3 lightDirection;

// is lightDirection actually an absolute position?
uniform bool isPosition;

// Phong shading parameters
uniform float diffuse;
uniform float specular;
uniform float shininess;

// forward declaration of mesh vertex shader functions (from Mesh.vert)
vec3 getVertex(sampler2D depthmap, vec2 position);
vec3 getVertexNormal(sampler2D depthmap, vec2 position, vec3 coords);

void main (void)
{
//---- diffuse and specular rendering -------------------------------------------------------------
	
	// vertex position and normal, in world coordinates
	vec2 pos = gl_FragCoord.xy - 1; // minus one to distribute border evenly on all sides
	vec3 vertex = getVertex(points, pos);
	vec3 normal = getVertexNormal(points, pos, vertex);
		
	// compute normal and vectors to light, viewer and reflected ray (all normalised)
	vec3 N = normalize(normal);
	vec3 L = normalize(lightDirection - (isPosition ? -vertex : vec3(0.0))); // translate if using absolute world coordinates
	vec3 V = vec3(0.0, 0.0, -1.0);
	vec3 R = normalize(-reflect(L, N));

	// diffuse tone image (ambient + diffuse terms)
	float diffuseTone = diffuse * max(dot(N, L), 0.0);
	
	// specular tone image (specular + shininess)
	float specularTone = specular * pow(max(dot(R, V), 0.0), shininess);

//---- computation of attribute maps --------------------------------------------------------------

	float detail = 1.0;
	switch(attributeMap)
	{
		case 1: // depth-based attribute map
			//// linear mapping
			//z = vec4((clamp(vertex.z, depthRange.x, depthRange.y) - depthRange.x) / (depthRange.y - depthRange.x));
			//z = vec4(clamp(vertex.z, depthRange.x, depthRange.y)) / 5;
			
			// per Barla et al. 2006
			detail = 1 - log(clamp(vertex.z, depthRange.x, depthRange.y) / depthRange.x) / log(depthRange.y / depthRange.x);
			break;
			
		case 2: // highlights-based attribute map
			detail = specularTone;
			break;
	}
	
//---- core X-Toon shader -------------------------------------------------------------------------

	// look up in toon texture
	gl_FragColor = texture2D(toonTexture, vec2(diffuseTone, detail));
	
	// extra debug outputs
	switch(attributeMap)
	{
		case 3: gl_FragColor = vec4(vec3(1 - log(clamp(vertex.z, depthRange.x, depthRange.y) / depthRange.x) / log(depthRange.y / depthRange.x)), 1); break;
		case 4: gl_FragColor = vec4(vec3(diffuseTone), 1); break;
		case 5: gl_FragColor = vec4(vec3(specularTone), 1); break;
	}
}
