// LibAvWrap -- .NET wrapper for the libav* libraries
// --------------------------------------------------
// By Christian Richardt <christian@richardt.name>
// (based on AvDn 0.7, www.informikon.com/tutorials/)
//
// $Id: FormatContext.cpp 436 2010-10-15 16:26:38Z cr333 $

#include "stdafx.h"
#include "ByteIOContext.h"
#include "CodecContext.h"
#include "FormatContext.h"
#include "OutputFormat.h"
#include "Packet.h"
#include "Stream.h"
#include "Utils.h"

namespace LibAvWrap
{
// constructors and destructor --------------------------------------------------------------------

	// Constructs an empty format context.
	LibAvFormatContext::LibAvFormatContext() : Wrapper<AVFormatContext>()
	{
		// av_alloc_format_context was renamed to avformat_alloc_context in revision 17047
		// and will be obsolete from version 53 of libavformat
		pointer = av_alloc_format_context();

		if(pointer == NULL)
			throw gcnew Exception("Could not allocate a format context.");

		need_to_free = true;
	}

	// Construct a format context from a given filename.
	LibAvFormatContext::LibAvFormatContext(String^ filename) : Wrapper<AVFormatContext>()
	{
		IntPtr ip = IntPtr::Zero;
		try
		{
			ip = Marshal::StringToHGlobalAnsi(filename);
			char* str = (char*)ip.ToPointer();
			AVFormatContext* tmpPtr;
			int result = av_open_input_file(&tmpPtr, str, NULL, 0, NULL);
			LibAvWrap::LibAvUtils::ThrowIfError(filename, result);
			pointer = tmpPtr;
			need_to_free = true;
		}
		finally
		{
			if(ip != IntPtr::Zero)
				Marshal::FreeHGlobal(ip);
		}
	}

	// Construct a format context from a given native pointer.
	LibAvFormatContext::LibAvFormatContext(IntPtr ptr) : Wrapper<AVFormatContext>(ptr)
	{
		need_to_free = false;
	}

	// Destructor.
	LibAvFormatContext::~LibAvFormatContext() 
	{
		if(need_to_free && pointer != NULL)
		{
			av_free(pointer);
			pointer = NULL;
		}
	}

// member functions -------------------------------------------------------------------------------

	// Closes the input file stream.
	void LibAvFormatContext::CloseInputFile()
	{
		av_close_input_file(pointer);
	}

	// Create a new stream.
	LibAvWrap::LibAvStream^ LibAvFormatContext::CreateNewStream(int nb)
	{
		AVStream* st = av_new_stream(pointer, nb);
		if(st == NULL) throw gcnew Exception("Could not allocate stream.");
		return gcnew LibAvStream(st);
	}

	// Dump some information about a format.
	void LibAvFormatContext::DumpFormat(int index, String^ url, bool isOutput)
	{
		IntPtr ip = Marshal::StringToHGlobalAnsi(url);
		char* str = (char*)ip.ToPointer();
		dump_format(pointer, index, str, (int)isOutput);
		Marshal::FreeHGlobal(ip);
	}

	// Returns the first video stream that is found.
	int LibAvFormatContext::FindFirstVideoStream()
	{
		for (int i = 0; i < NbStreams; i++)
			if (GetCodecTypeForStream(i) == LibAvWrap::MediaType::Video)
				return i;
		return -1;
	}

	// Returns the stream info.
	int LibAvFormatContext::FindStreamInfo()
	{
		return av_find_stream_info(pointer);
	}

	LibAvCodecContext^ LibAvFormatContext::GetCodecContextForStream(int i)
	{
		if(i < 0 || i > (int)pointer->nb_streams)
			return gcnew LibAvCodecContext((AVCodecContext*)NULL);
		AVCodecContext* cCtx = pointer->streams[i]->codec;
		return gcnew LibAvCodecContext(cCtx);
	}

	// Returns the codec type for a stream.
	LibAvWrap::MediaType LibAvFormatContext::GetCodecTypeForStream(int i)
	{
		if(i < 0 || i > (int)pointer->nb_streams)
			return LibAvWrap::MediaType::Unknown;
		return (LibAvWrap::MediaType)pointer->streams[i]->codec->codec_type;
	}

	// Returns a pointer to "the bytes" as a ByteIOContext.
	LibAvByteIOContext^ LibAvWrap::LibAvFormatContext::GetPb()
	{
		return gcnew LibAvByteIOContext(IntPtr((void*)&pointer->pb));
	}

	// Reads a frame from a packet.
	int LibAvFormatContext::ReadFrame(LibAvPacket^ p)
	{
		return av_read_frame(pointer, p->Pointer);
	}

	// Set some parameters (use with null only so far).
	int LibAvFormatContext::SetParameters(IntPtr ptr)
	{
		AVFormatParameters* pFormatParams = (AVFormatParameters*)(void*)ptr ; 
		return av_set_parameters(pointer, pFormatParams);
		// TODO: use ThrowIfError on the return value
	}

	// Writes a frame from a packet.
	int LibAvFormatContext::WriteFrame(LibAvPacket^ p)
	{
		// av_interleaved_write_frame(oc, &pkt); // cr333: in latest code
		return av_write_frame(pointer, p->Pointer);
	}

	// Write the header for a stream given the current AVFormatContext.
	void LibAvFormatContext::WriteHeader()
	{
		av_write_header(pointer);
	}

	// Write the trailer for a stream given the current AVFormatContext.
	void LibAvFormatContext::WriteTrailer()
	{
		av_write_trailer(pointer);
	}
	
// properties -------------------------------------------------------------------------------------

	// Get the input or output filename.
	String^ LibAvFormatContext::Filename::get()
	{
		return Marshal::PtrToStringAnsi(IntPtr((void*)(pointer->filename)));
	}

	// Set the input or output filename.
	void LibAvFormatContext::Filename::set(String^ filename)
	{
		// have to marshal filename
		_snprintf_s(pointer->filename, sizeof(pointer->filename), "%s", filename);
	}

	// Get the number of streams field of the AVFormatContext structure.
	int LibAvFormatContext::NbStreams::get()
	{
		return pointer->nb_streams;
	}

	// Output format field of the AVFormatContext structure.
	LibAvOutputFormat^ LibAvFormatContext::OutputFormat::get()
	{
		if(pointer->oformat == NULL) return nullptr;
		else return gcnew LibAvOutputFormat(pointer->oformat);
	}

	// Output format field of the AVFormatContext structure.
	void LibAvFormatContext::OutputFormat::set(LibAvOutputFormat^ format)
	{
		AVOutputFormat* pOutputFormat = format->Pointer;
		pointer->oformat = pOutputFormat;
	}

} // namespace LibAvWrap
