﻿// $Id: LibAvConvert.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.Text;
using VideoLib.Frames;
using LibAvWrap;
using System.Drawing;

namespace VideoLib.LibAv
{
	/// <summary>Defines how frames are resampled.</summary>
	[Flags] public enum ResamplingMode
	{
		/// <summary>Nearest-neighbour interpolation is the fastest, but roughest technique.</summary>
		NearestNeighbour = SwsFlags.Point,

		/// <summary>Bilinear interpolation provides a good trade-off between speed and accuracy.</summary>
		Bilinear = SwsFlags.Bilinear,

		/// <summary>Bicubic interpolation provides the best interpolation results available here.</summary>
		Bicubic = SwsFlags.Bicubic,

		/// <summary>A faster bilinear resampling mode.</summary>
		FastBilinear = SwsFlags.FastBilinear,

		/// <summary>"Experimental scaler".</summary>
		X = SwsFlags.X,

		/// <summary>Area averaging scaler.</summary>
		Area = SwsFlags.Area,

		/// <summary>Luma bicubic, chroma bilinear scaler.</summary>
		Bicublin = SwsFlags.Bibublin,

		/// <summary>Gaussian scaler.</summary>
		Gauss = SwsFlags.Gauss,

		/// <summary>Sinc scaler.</summary>
		Sinc = SwsFlags.Sinc,

		/// <summary>Bicubic spline scaler.</summary>
		Spline = SwsFlags.Spline,

		/// <summary>Lanczos scaler.</summary>
		Lanczos = SwsFlags.Lanczos,
	}

	/// <summary>Fast bitmap converter and scaler based on libswscale.</summary>
	public class LibAvConvert : Node
	{
		/// <summary>Target output format. Default: Unknown, i.e. no format conversion.</summary>
		public VideoLib.Frames.PixelFormat OutFormat { get; set; }

		/// <summary>Target width.</summary>
		public int Width { get; set; }

		/// <summary>Target height.</summary>
		public int Height { get; set; }

		/// <summary>Target size.</summary>
		public Size Size { set { Width = value.Width; Height = value.Height; } }

		/// <summary>Interpolation technique to use (default: fast bilinear).</summary>
		public ResamplingMode Mode { get; set; }

		/// <summary>If true, will try to preserve the aspect ratio of the input frames. Default: true.</summary>
		public bool PreserveAspectRatio { get; set; }

		/// <summary>Human-readable name of this node.</summary>
		public override string Name
		{
			get
			{
				if (OutFormat != VideoLib.Frames.PixelFormat.Unknown)
				{
					if (Width == 0 && Height == 0)
						return String.Format("{0}({1})", base.Name, OutFormat);
					else
						return String.Format("{0}({1}x{2}, {3})", base.Name, Width, Height, OutFormat);
				}
				else // output format same as input format
				{
					if (Width == 0 && Height == 0)
						return String.Format("{0}(???)", base.Name);
					else
						return String.Format("{0}({1}x{2})", base.Name, Width, Height);
				}
			}
		}


		/// <summary>Constructs a no-op frame conversion/scaling node that can be individually configured.</summary>
		public LibAvConvert()
			: base()
		{
			AddInputPin("in");
			AddOutputPin("out");
			OutFormat = VideoLib.Frames.PixelFormat.Unknown;
			Mode = ResamplingMode.FastBilinear;
			PreserveAspectRatio = true;
		}

		/// <summary>Constructs a scaling node for the given target resolution.</summary>
		/// <param name="targetSize">Target output size.</param>
		public LibAvConvert(Size targetSize) : this() { Size = targetSize; }

		/// <summary>Constructs a format conversion node for the given format.</summary>
		/// <param name="targetFormat">Target frame format.</param>
		public LibAvConvert(VideoLib.Frames.PixelFormat targetFormat) : this() { OutFormat = targetFormat; }

		/// <summary>Processes a video frame.</summary>
		public override void Process()
		{
			// get the input BitmapFrame
			Frame frame = GetInputFrame("in");
			if (PassOnEndOfStreamFrame(frame)) return;
			BitmapFrame bitmapFrame = CastFrameTo<BitmapFrame>(frame);

			// get input format and desired output format
			LibAvWrap.PixelFormat srcFormat = bitmapFrame.Format.ToLibAvPixelFormat();
			LibAvWrap.PixelFormat dstFormat = OutFormat.ToLibAvPixelFormat();

			// if not defined, use same format as source
			if (OutFormat == VideoLib.Frames.PixelFormat.Unknown)
				dstFormat = srcFormat;

			// check for valid formats
			if (srcFormat == LibAvWrap.PixelFormat.None || dstFormat == LibAvWrap.PixelFormat.None)
				throw new NotImplementedException();

			// compute target size
			int targetWidth;
			int targetHeight;
			ComputeTargetSize(bitmapFrame, out targetWidth, out targetHeight);

			BitmapFrame outputFrame = null;

			// take shortcuts if same size as input frame
			if (bitmapFrame.Width == targetWidth && bitmapFrame.Height == targetHeight)
			{
				// skip if no change required
				if (srcFormat == dstFormat)
				{
					outputFrame = (BitmapFrame)bitmapFrame;
				}
				else // just need to change format, not size
				{
					outputFrame = ConvertFrameFormat(bitmapFrame, srcFormat, dstFormat);
				}
			}
			else // need to scale and maybe convert format
			{
				outputFrame = ScaleFrame(bitmapFrame, srcFormat, dstFormat, targetWidth, targetHeight);
			}

			outputFrame.SequenceNumber = bitmapFrame.SequenceNumber;
			PushFrame("out", outputFrame);
		}

		/// <summary>Fits the BitmapFrame into the target width/height.</summary>
		/// <param name="bitmapFrame">The bitmap frame.</param>
		/// <param name="targetWidth">Target Width.</param>
		/// <param name="targetHeight">Target height.</param>
		private void ComputeTargetSize(BitmapFrame bitmapFrame, out int targetWidth, out int targetHeight)
		{
			// for the output, use the desired size or the same as the input frame
			targetWidth = (Width != 0 ? Width : bitmapFrame.Width);
			targetHeight = (Height != 0 ? Height : bitmapFrame.Height);

			if (PreserveAspectRatio)
			{
				// fit to width if input image is wider than output image
				if (bitmapFrame.Width * targetHeight >= targetWidth * bitmapFrame.Height) // multiplied out numerator
					targetHeight = (int)(0.5 + (float)bitmapFrame.Height / bitmapFrame.Width * targetWidth);
				else // fit to height if input image is higher than output image
					targetWidth = (int)(0.5 + (float)bitmapFrame.Width / bitmapFrame.Height * targetHeight);
			}
		}

		/// <summary>Converts a BitmapFrame to a different format.</summary>
		/// <param name="bitmapFrame">The bitmap frame.</param>
		/// <param name="srcFormat">The source format.</param>
		/// <param name="dstFormat">The target format.</param>
		/// <returns>The converted BitmapFrame.</returns>
		private static BitmapFrame ConvertFrameFormat(BitmapFrame bitmapFrame, LibAvWrap.PixelFormat srcFormat, LibAvWrap.PixelFormat dstFormat)
		{
			// wrap input frame in a LibAvFrame
			LibAvFrame srcFrame = bitmapFrame.ToLibAvFrame();

			// prepare a new empty frame to write to
			LibAvFrame frameRGB = new LibAvFrame(dstFormat, bitmapFrame.Width, bitmapFrame.Height);

			// format conversion
			srcFrame.ConvertTo(frameRGB);

			// convert back to BitmapFrame and clean up
			BitmapFrame outputFrame = frameRGB.ToBitmapFrame();
			frameRGB.Dispose();
			srcFrame.Dispose();
			return outputFrame;
		}

		/// <summary>Scales (and converts) a BitmapFrame.</summary>
		/// <param name="bitmapFrame">The bitmap frame.</param>
		/// <param name="srcFormat">The source format.</param>
		/// <param name="dstFormat">The target format.</param>
		/// <param name="targetWidth">Target width.</param>
		/// <param name="targetHeight">Target Height.</param>
		/// <returns>The scaled BitmapFrame.</returns>
		private BitmapFrame ScaleFrame(BitmapFrame bitmapFrame, LibAvWrap.PixelFormat srcFormat, LibAvWrap.PixelFormat dstFormat, int targetWidth, int targetHeight)
		{
			// wrap input frame in a LibAvFrame
			LibAvFrame srcFrame = bitmapFrame.ToLibAvFrame();

			// prepare a new empty frame to write to
			LibAvFrame frameRGB = new LibAvFrame(dstFormat, targetWidth, targetHeight);

			// apply the scaling or format coversion
			LibAvFrame.Scale(srcFrame, frameRGB, (SwsFlags)Mode);

			// convert back to BitmapFrame and clean up
			BitmapFrame outputFrame = frameRGB.ToBitmapFrame();
			frameRGB.Dispose();
			srcFrame.Dispose();

			return outputFrame;
		}
	}
}
