﻿// $Id: LibAvSource.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.Text;
using VideoLib.Frames;
using System.Threading;
using System.Runtime.InteropServices;
using System.IO;
using System.Diagnostics;
using LibAvWrap;

namespace VideoLib.LibAv
{
	/// <summary>A video source based on libavcodec.</summary>
	public class LibAvSource : Node
	{
		/// <summary>The filename or URL of the underlying video.</summary>
		public string Filename { get; protected set; }

		/// <summary>The sequence number of the most recently decoded frame.</summary>
		public int SequenceNumber { get; protected set; }

		/// <summary>Skips the first n frames. Default: 0.</summary>
		public int SkipFirstFrames { get; set; }

		/// <summary>The framerate with which the video frames will be output. Default: 10.</summary>
		public float FPS { get; set; }

		/// <summary>Human-friendly node name.</summary>
		public override string Name
		{
			get { return String.Format("{0}({1})", base.Name, Path.GetFileName(Filename)); }
		}

		/// <summary>Sets or gets the desired frame format type. Default: BGR24.</summary>
		public VideoLib.Frames.PixelFormat Format { get; set; }

		private bool ready = false;
		private bool finished = false;

		// state required for decoding a video
		private LibAvFormatContext formatContext = null;
		private int videoStream = -1;
		private LibAvCodecContext codecContext = null;
		private DateTime lastFrame = DateTime.Now;

		/// <summary>Initialises all LibAvCodecSources.</summary>
		static LibAvSource()
		{
			// Initialize libavformat and register all the muxers, demuxers and protocols.
			LibAvUtils.RegisterAll();
			LibAvUtils.UsePrintfLogger();
		}

		/// <summary>Creates a video source based on libavcodec.</summary>
		/// <param name="filename">Video filename or URL.</param>
		public LibAvSource(string filename)
			: base()
		{
			Filename = filename;
			SequenceNumber = 0;
			SkipFirstFrames = 0;
			FPS = 10;
			Format = VideoLib.Frames.PixelFormat.BGR24;
			AddOutputPin("out");
		}

		/// <summary>Initialises the video decoder.</summary>
		public override void Init()
		{
			formatContext = new LibAvFormatContext(Filename);
			if (formatContext.IntPointer == IntPtr.Zero || formatContext.FindStreamInfo() < 0)
				throw new Exception("Couldn't find stream info");

			videoStream = formatContext.FindFirstVideoStream();
			if (videoStream == -1)
				throw new Exception("couldn't find video stream");

			codecContext = formatContext.GetCodecContextForStream(videoStream);
			if (codecContext.OpenDecoder() < 0)
				throw new Exception("couldn't open decoder");

			ready = true;
		}

		/// <summary>Update this node's status.</summary>
		protected override void UpdateStatusImpl()
		{
			// don't update if currently running
			if (Status == NodeStatus.Running)
				return;

			// if ready, but not finished, and the output pin is ready
			if (ready && !finished && OutputPins["out"].Status == PinStatus.Ready)
			{
				Status = NodeStatus.Ready;
				return;
			}

			// once finished, stay in that state
			if (finished)
				Status = NodeStatus.Finished;

			// ready? give me data!
			if (Status == NodeStatus.Ready)
				return;

			// otherwise: wait for downstream processing to unblock
			Status = NodeStatus.Waiting;
		}

		/// <summary>Decode a single frame from the current video.</summary>
		public override void Process()
		{
			// keep reading in frames until we have a complete video frame
			LibAvPacket packet = new LibAvPacket();
			while (formatContext.ReadFrame(packet) >= 0)
			{
				// ignore audio frames
				if (packet.StreamIndex != videoStream)
					continue;

				// Now that we've got a video frame, let's try to decode it.
				// DecodeVideo will store the decoded video frame in <frame>.
				int gotPicture;
				LibAvFrame frame = new LibAvFrame();
				codecContext.DecodeVideo(out frame, packet, out gotPicture);
				packet.Free();

				// keep going until the frame has been successfully decoded
				if (gotPicture == 0)
					continue;

				// skip first few frames if asked to do so
				if (SequenceNumber < SkipFirstFrames)
				{
					SequenceNumber++;
					continue;
				}

				// prepare a new empty frame to write to
				LibAvFrame targetFrame = new LibAvFrame(Format.ToLibAvPixelFormat(), frame.Width, frame.Height);
				frame.ConvertTo(targetFrame);

				// prepare a BitmapFrame to send on
				BitmapFrame bitmapFrame = targetFrame.ToBitmapFrame();
				bitmapFrame.SequenceNumber = SequenceNumber;
				SequenceNumber++;
				targetFrame.Dispose();

				// send on frame
				if(Verbose) Debug.WriteLine(String.Format("{0}: decoded frame #{1} at libavsource", Thread.CurrentThread.Name, SequenceNumber - 1));
				GetOutputPin("out").PushFrame(bitmapFrame);

				// sleep to get framerate right
				// TODO: use StopWatch or some other high-precision timer
				//       (Sleep only has a resolution of roughly 10ms)
				Thread.Sleep((int)Math.Max(0, 1000 / FPS - (DateTime.Now.Ticks - lastFrame.Ticks) / 1e4));
				lastFrame = DateTime.Now;
				return;
			}

			// no more data, clean up
			CleanUp();

			// send end-of-stream frame
			EndOfStreamFrame eosFrame = new EndOfStreamFrame() { SequenceNumber = SequenceNumber };
			GetOutputPin("out").PushFrame(eosFrame);
			Status = NodeStatus.Finished;
		}

		/// <summary>close the currently open file and clean up things.</summary>
		protected void CleanUp()
		{
			ready = false;
			finished = true;
			if (codecContext != null) { codecContext.Close(); codecContext = null; }
			if (formatContext != null) { formatContext.CloseInputFile(); formatContext = null; }
		}

		/// <summary>CleanUp up open files etc.</summary>
		protected override void Dispose(bool disposing)
		{
			CleanUp();
			base.Dispose(disposing);
		}
	}
}
