﻿// $Id: JointBilateralFilter.cs 499 2010-11-17 16:29:28Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;

namespace VideoLib.OpenGL
{
	/// <summary>Applies a joint-bilateral filter that filters values while preserving edges in the edge image.</summary>
	public class JointBilateralFilter : FboNode
	{
		/// <summary>Gets or sets the spatial sigma.</summary>
		[Uniform(), Description("Standard deviation for spatial filtering.")]
		public float SpatialSigma { get; set; }

		/// <summary>Gets or sets the range sigma.</summary>
		[Uniform(), Description("Standard deviation for range filtering.")]
		public float RangeSigma { get; set; }

		/// <summary>Initialises a new instance of the <see cref="BilateralFilter"/> class.</summary>
		public JointBilateralFilter()
		{
			// default settings
			SpatialSigma = 4;
			RangeSigma   = 0.1f;

			AddInputPins("edges", "values");
			AddOutputPin("out");
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.JointBilateralFilter.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Name = "edges",  Unit = 0 },
				new TextureBinding { Name = "values", Unit = 1 },
			};

			base.Init();
		}

		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			AssertFrameProperties(inputs, FrameProps.Size);

			// allocate a texture that is identical to the input
			var resultTexture = TextureFrame.CreateToMatch(inputs["values"]);
			resultTexture.Alloc();

			Draw(inputs, resultTexture.Texture);

			return resultTexture;
		}
	}
}
