﻿// $Id: TextureFrame.cs 495 2010-11-15 20:32:40Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;
using System.Threading;
using OpenTK.Graphics.OpenGL;

namespace VideoLib.OpenGL
{
	/// <summary>A video frame based on an OpenGL texture.</summary>
	public class TextureFrame : VideoFrame, IRefCountedFrame
	{
		/// <summary>Gets or sets the texture contained in this frame.</summary>
		public Texture Texture { get; protected set; }

		#region Proxy interface
//=================================================================================================

		/// <summary>Gets or sets the internal pixel format of the texture. Default: RGBA16F.</summary>
		public PixelInternalFormat InternalFormat
		{
			get { return Texture.InternalFormat; }
			set { Texture.InternalFormat = value; }
		}

		/// <summary>Gets or sets the texture magnification filter of the texture. Default: linear.</summary>
		public TextureMagFilter MagFilter
		{
			get { return Texture.MagFilter; }
			set { Texture.MagFilter = value; }
		}

		/// <summary>Gets or sets the texture minification filter of the texture. Default: linear.</summary>
		public TextureMinFilter MinFilter
		{
			get { return Texture.MinFilter; }
			set { Texture.MinFilter = value; }
		}

		/// <summary>Sets both minification and magnification filters of the texture. Write-only.</summary>
		public TextureMagFilter Filter
		{
			set { MagFilter = value; MinFilter = (TextureMinFilter)(int)value; }
		}

		/// <summary>Gets or sets the horizontal texture wrap mode of the texture. Default: clamp.</summary>
		public TextureWrapMode WrapS
		{
			get { return Texture.WrapS; }
			set { Texture.WrapS = value; }
		}

		/// <summary>Gets or sets the vertical texture wrap mode of the texture. Default: clamp.</summary>
		public TextureWrapMode WrapT
		{
			get { return Texture.WrapT; }
			set { Texture.WrapT = value; }
		}

		/// <summary>Sets both texture wrap modes of the texture. Write-only.</summary>
		public TextureWrapMode Wrap
		{
			set { WrapS = value; WrapT = value; }
		}

		/// <summary>Allocates an empty texture according to the configured parameters.</summary>
		/// <returns>Texture ID, for convenience.</returns>
		public int Alloc() { return Texture.Alloc(); }
		
		/// <summary>Binds the underlying texture to the current texture unit.</summary>
		public void Bind() { Texture.Bind(); }

		/// <summary>Binds the underlying texture to the given texture unit.</summary>
		/// <param name="unit">The texture unit.</param>
		public void Bind(TextureUnit unit) { Texture.Bind(unit); }

		/// <summary>Unbinds the underlying texture.</summary>
		public void Unbind() { Texture.Unbind(); }

//=================================================================================================
		#endregion

		/// <summary>Initializes the <see cref="TextureFrame"/> class.</summary>
		static TextureFrame()
		{
			// use default handler for texture deletion
			DeleteTextureHandler = DefaultDeleteTextureHandler;
		}

		/// <summary>Initialises a new instance of the <see cref="TextureFrame"/> class of the given size.</summary>
		/// <param name="width">The width of the texture.</param>
		/// <param name="height">The height of the texture.</param>
		public TextureFrame(int width, int height)
		{
			refCount   = 1;
			needToFree = true;

			Width   = width;
			Height  = height;
			Texture = new Texture(width, height);
		}

		/// <summary>Initialises a new instance of the <see cref="TextureFrame"/> class from a given texture ID.</summary>
		/// <param name="textureId">The texture ID.</param>
		public TextureFrame(int textureId)
		{
			refCount   = 1;
			needToFree = false;

			Texture = new Texture(textureId);
			Width   = Texture.Width;
			Height  = Texture.Height;
		}

		/// <summary>Performs an implicit conversion from <see cref="TextureFrame"/> to <see cref="Texture"/>.</summary>
		/// <param name="frame">The input texture frame.</param>
		/// <returns>The texture of the frame.</returns>
		public static implicit operator Texture(TextureFrame frame)
		{
			return frame.Texture;
		}

		/// <summary>Creates a new <see cref="TextureFrame"/> that matches the given one in every aspect except contents.</summary>
		/// <param name="other">The model <see cref="TextureFrame"/>.</param>
		/// <returns>New <see cref="TextureFrame"/> that matches the input.</returns>
		public static TextureFrame CreateToMatch(TextureFrame other)
		{
			return new TextureFrame(other.Width, other.Height)
			{
				SequenceNumber = other.SequenceNumber,
				InternalFormat = other.InternalFormat,
				MinFilter      = other.MinFilter,
				MagFilter      = other.MagFilter,
				WrapS          = other.WrapS,
				WrapT          = other.WrapT
			};
		}

		#region Reference counting
//=================================================================================================

		/// <summary>Gets the reference count of this texture frame.</summary>
		public int RefCount { get { return refCount; } }
		private int refCount;

		private bool needToFree;

		/// <summary>Delegate type for deleting textures.</summary>
		public delegate void DeleteTextureDelegate(int textureId);

		/// <summary>Delegate for deleting textures on the OpenGL thread.</summary>
		public static DeleteTextureDelegate DeleteTextureHandler;

		/// <summary>Default handler for texture deletion: just free it (in current thread).</summary>
		/// <param name="textureID">The texture ID.</param>
		public static void DefaultDeleteTextureHandler(int textureID)
		{
			// Note:
			// This causes an AccessViolationException if called from the finaliser thread,
			// as textures need to be freed on the OpenGL thread, not the finaliser thread.
			GL.DeleteTexture(textureID);
		}

		/// <summary>
		/// The finaliser releases unmanaged resources and performs other cleanup operations
		/// before the <see cref="Texture"/> is reclaimed by garbage collection.
		/// </summary>
		~TextureFrame()
		{
			if (refCount > 0 && needToFree)
			{
#if DEBUG
				Console.WriteLine(String.Format("Warning: Referenced {0} is finalised.", this));
				Console.WriteLine(String.Format("         The reference count was {0}, memory leak possible.", RefCount));
#endif

				// free texture on the OpenGL thread
				DeleteTextureHandler(Texture);
			}
		}

		/// <summary>Increments the reference count of this frame.</summary>
		public void RefInc()
		{
			Interlocked.Increment(ref refCount);
		}

		/// <summary>Decrements the reference count of this frame.</summary>
		/// <remarks>If the reference count reaches zero, the memory is deallocated.</remarks>
		public void RefDec()
		{
			if (Interlocked.Decrement(ref refCount) == 0)
			{
				if (Texture != null && needToFree)
				{
					// free texture on the OpenGL thread
					DeleteTextureHandler(Texture);
					Texture = null;
				}
			}
		}

//=================================================================================================
		#endregion // Reference counting

		/// <summary>Copy this frame's contents.</summary>
		/// <param name="newFrame">The destination Frame where this Frame's data should be copied to.</param>
		/// <remarks>All frames should be able to copy their contents to another Frame.</remarks>
		public override void copyTo(ref Frame newFrame)
		{
			throw new NotImplementedException();
		}

		/// <summary>Returns a <see cref="System.String"/> that represents this instance.</summary>
		public override string ToString()
		{
			return String.Format("{0} ({1}x{2}, #{3})", this.GetType().Name, Width, Height, Texture.ID);
		}
	}
}
