﻿// $Id: AlphaExtractor.cs 262 2010-07-28 13:25:26Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib;
using VideoLib.Frames;

namespace VideoLib.Filters
{
	/// <summary>
	/// A simple node that extracts the alpha channel of an BGRA32 frame and outputs it as a Gray8.
	/// </summary>
	public class AlphaExtractor : Node
	{
		/// <summary>
		/// Creates a new instance of the alpha extractor node.
		/// </summary>
		public AlphaExtractor()
			: base()
		{
			AddInputPin("in");
			AddOutputPin("out");
		}

		/// <summary>
		/// Processes some frames, to produce new frames.
		/// </summary>
		public override void Process()
		{
			// get input frame
			Frame frame = GetInputFrame("in");
			if (PassOnEndOfStreamFrame(frame)) return;
			BitmapFrame inputFrame = CastFrameTo<BitmapFrame>(frame);

			// extract the alpha channel
			BitmapFrame outputFrame = ExtractAlpha(inputFrame);
			PushFrame(outputFrame);
		}

		/// <summary>
		/// Extracts the input frame's alpha channel and returns it as a Grey8 bitmap frame.
		/// </summary>
		/// <param name="inputFrame">The input frame. Needs to be BGRA32.</param>
		/// <returns>The alpha channel as a Grey8 frame.</returns>
		private unsafe BitmapFrame ExtractAlpha(BitmapFrame inputFrame)
		{
			AssertPixelFormat(inputFrame, PixelFormat.BGRA32);
			BitmapFrame outputFrame = new BitmapFrame(inputFrame.Width, inputFrame.Height, PixelFormat.Grey8);
			outputFrame.SequenceNumber = inputFrame.SequenceNumber;

			fixed (byte* dataI = inputFrame.Data, dataO = outputFrame.Data)
			{
				for (int y = 0; y < outputFrame.Height; y++)
				{
					int offsetI = y * inputFrame.Stride;
					int offsetO = y * outputFrame.Stride;

					for (int x = 0; x < outputFrame.Width; x++)
					{
						dataO[offsetO + x] = dataI[offsetI + 4 * x + 3];
					}
				}
			}

			return outputFrame;
		}
	}
}
