﻿// $Id: AnaglyphFilter.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;

namespace VideoLib.Filters
{
	/// <summary>Describes a particular kind of anaglyph method.</summary>
	public enum AnaglyphMethod
	{
		/// <summary>True anaglyph: dark image and no colour reproduction, but little ghosting.</summary>
		True,

		/// <summary>Grey anaglyph: no colour reproduction and more ghosting than true anaglyphs.</summary>
		Grey,

		/// <summary>Colour anaglyphs: partial colour reproduction, but retinal rivalry.</summary>
		Colour,

		/// <summary>Half colour anaglyph: partial colour reproduction, but less retinal rivalry than colour anaglyphs.</summary>
		HalfColour
	}

	/// <summary>Anaglgyph filter that overleys two images to be seen with red/cyan glasses.</summary>
	public class AnaglyphFilter : Node
	{
		/// <summary>Anaglyph method to use.</summary>
		public AnaglyphMethod Method { get; set; }

		/// <summary>Constructs a new anaglyph filter.</summary>
		/// <param name="method">Anaglyph method to use.</param>
		public AnaglyphFilter(AnaglyphMethod method) : base()
		{
			Method = method;
			AddInputPins("inL", "inR");
			AddOutputPin("out");
		}

		/// <summary>
		/// Returns a <see cref="System.String"/> that represents this instance.
		/// </summary>
		public override string ToString()
		{
			return String.Format("Anaglyph({0})", Method);
		}

		/// <summary>Processes a frame.</summary>
		public override void Process()
		{
			// get input frames
			Frame inL = GetInputFrame("inL");
			Frame inR = GetInputFrame("inR");
			if (PassOnEndOfStreamFrame(inL) || PassOnEndOfStreamFrame(inR)) return;

			// get the input bitmap data
			if(!(inL is BitmapFrame)) throw new FrameNotSupportedException(inL);
			if(!(inR is BitmapFrame)) throw new FrameNotSupportedException(inR);
			BitmapFrame bitmapL = inL as BitmapFrame;
			BitmapFrame bitmapR = inR as BitmapFrame;

			// check for same size
			if (bitmapL.Width != bitmapR.Width || bitmapL.Height != bitmapR.Height)
				throw new Exception("Both input frames must have same size.");

			// check for same format
			if (bitmapL.Format != bitmapR.Format)
				throw new Exception("Both input frames must have same format.");

			// check for same sequence number
			if (bitmapL.SequenceNumber != bitmapR.SequenceNumber)
				throw new Exception("Both input frames must have same sequence number.");

			// look up the colour mixing for the particular method
			ColourMixing mix = GetColourMixing(Method);

			// apply the anaglyph impementation for the given format
			BitmapFrame outputFrame = null;
			switch (bitmapL.Format)
			{
				case PixelFormat.Grey8: outputFrame = ComputeAnaglyphGrey8(bitmapL, bitmapR, mix); break;
				case PixelFormat.BGR24: outputFrame = ComputeAnaglyphBgr24(bitmapL, bitmapR, mix); break;
				case PixelFormat.BGR32: outputFrame = ComputeAnaglyphBgr32(bitmapL, bitmapR, mix); break;

				default:
					throw new NotImplementedException(String.Format("Bitmap format \"{0}\" is not supported. Try BGR24 or BGR32.", bitmapL.Format));
			}

			// output the frame
			PushFrame("out", outputFrame);
		}

		#region Colour Mixing

		/// <summary>Describes the colour mixing for some anaglyph method.</summary>
		private class ColourMixing
		{
			/// <summary>Gets or sets the left image colour mixing matrix.</summary>
			/// <remarks>The 3 rows are for the RGB outputs, the 3 columns for the RGB inputs. Example: [3] is the red-to-green coeeficient.</remarks>
			public float[] CoeffsL { get; set; }

			/// <summary>Gets or sets the right image colour mixing matrix.</summary>
			/// <remarks>The 3 rows are for the RGB outputs, the 3 columns for the RGB inputs. Example: [3] is the red-to-green coeeficient.</remarks>
			public float[] CoeffsR { get; set; }

			/// <summary>Initializes a new instance of the <see cref="ColourMixing"/> class.</summary>
			public ColourMixing()
			{
				// default mix: black
				CoeffsL = new float[] { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
				CoeffsR = new float[] { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
			}
		}

		/// <summary>Returns the colour mixing according to http://www.3dtv.at/Knowhow/AnaglyphComparison_en.aspx.</summary>
		/// <param name="Method">The anaglyph method.</param>
		private ColourMixing GetColourMixing(AnaglyphMethod Method)
		{
			switch (Method)
			{
				case AnaglyphMethod.True:
					return new ColourMixing()
					{
						CoeffsL = new float[] { 0.299f, 0.587f, 0.114f, 0, 0, 0, 0, 0, 0 },
						CoeffsR = new float[] { 0, 0, 0, 0, 0, 0, 0.299f, 0.587f, 0.114f }
					};

				case AnaglyphMethod.Grey:
					return new ColourMixing()
					{
						CoeffsL = new float[] { 0.299f, 0.587f, 0.114f, 0, 0, 0, 0, 0, 0 },
						CoeffsR = new float[] { 0, 0, 0, 0.299f, 0.587f, 0.114f, 0.299f, 0.587f, 0.114f }
					};

				case AnaglyphMethod.Colour:
					return new ColourMixing()
					{
						CoeffsL = new float[] { 1, 0, 0, 0, 0, 0, 0, 0, 0 },
						CoeffsR = new float[] { 0, 0, 0, 0, 1, 0, 0, 0, 1 }
					};

				case AnaglyphMethod.HalfColour:
					return new ColourMixing()
					{
						CoeffsL = new float[] { 0.299f, 0.587f, 0.114f, 0, 0, 0, 0, 0, 0 },
						CoeffsR = new float[] { 0, 0, 0, 0, 1, 0, 0, 0, 1 }
					};

				default:
					throw new Exception("Unknown anaglyph method.");
			}
		}

		#endregion

		#region Anaglyph Implementations

		/// <summary>Computes an anaglyph images from 8-bit greyscale images.</summary>
		/// <param name="bitmapL">Left input frame.</param>
		/// <param name="bitmapR">Right input frame.</param>
		/// <param name="mix">The colour mixing for the selected anaglyph method.</param>
		/// <returns>The anaglyph image, packaged as a BitmapFrame.</returns>
		private unsafe BitmapFrame ComputeAnaglyphGrey8(BitmapFrame bitmapL, BitmapFrame bitmapR, ColourMixing mix)
		{
			BitmapFrame outputFrame = new BitmapFrame(bitmapL.Width, bitmapL.Height, PixelFormat.BGR24) { SequenceNumber = bitmapL.SequenceNumber };

			fixed (byte* dataL = bitmapL.Data, dataR = bitmapR.Data, dataO = outputFrame.Data)
			{
				for (int y = 0; y < outputFrame.Height; y++)
				{
					int offsetL = y * bitmapL.Stride;
					int offsetR = y * bitmapR.Stride;
					int offsetO = y * outputFrame.Stride;

					for (int x = 0; x < outputFrame.Width; x++)
					{
						float pixL = dataL[offsetL + x];
						float pixR = dataR[offsetR + x];
						dataO[offsetO + 3 * x + 2] = (byte)((mix.CoeffsL[0] + mix.CoeffsL[1] + mix.CoeffsL[2]) * pixL + (mix.CoeffsR[0] + mix.CoeffsR[1] + mix.CoeffsR[2]) * pixR);
						dataO[offsetO + 3 * x + 1] = (byte)((mix.CoeffsL[3] + mix.CoeffsL[4] + mix.CoeffsL[5]) * pixL + (mix.CoeffsR[3] + mix.CoeffsR[4] + mix.CoeffsR[5]) * pixR);
						dataO[offsetO + 3 * x + 0] = (byte)((mix.CoeffsL[6] + mix.CoeffsL[7] + mix.CoeffsL[8]) * pixL + (mix.CoeffsR[6] + mix.CoeffsR[7] + mix.CoeffsR[8]) * pixR);
					}
				}
			}

			return outputFrame;
		}

		/// <summary>Computes an anaglyph images from BGR24 images.</summary>
		/// <param name="bitmapL">Left input frame.</param>
		/// <param name="bitmapR">Right input frame.</param>
		/// <param name="mix">The colour mixing for the selected anaglyph method.</param>
		/// <returns>The anaglyph image, packaged as a BitmapFrame.</returns>
		private unsafe BitmapFrame ComputeAnaglyphBgr24(BitmapFrame bitmapL, BitmapFrame bitmapR, ColourMixing mix)
		{
			BitmapFrame outputFrame = new BitmapFrame(bitmapL.Width, bitmapL.Stride, bitmapL.Height, bitmapL.Format) { SequenceNumber = bitmapL.SequenceNumber };

			fixed (byte* dataL = bitmapL.Data, dataR = bitmapR.Data, dataO = outputFrame.Data)
			{
				for (int y = 0; y < outputFrame.Height; y++)
				{
					int offsetL = y * bitmapL.Stride;
					int offsetR = y * bitmapR.Stride;
					int offsetO = y * outputFrame.Stride;

					for (int x = 0; x < 3 * outputFrame.Width; x += 3)
					{
						dataO[offsetO + x + 2] = (byte)(mix.CoeffsL[0] * dataL[offsetL + x + 2] + mix.CoeffsL[1] * dataL[offsetL + x + 1] + mix.CoeffsL[2] * dataL[offsetL + x + 0] + mix.CoeffsR[0] * dataR[offsetR + x + 2] + mix.CoeffsR[1] * dataR[offsetR + x + 1] + mix.CoeffsR[2] * dataR[offsetR + x + 0]);
						dataO[offsetO + x + 1] = (byte)(mix.CoeffsL[3] * dataL[offsetL + x + 2] + mix.CoeffsL[4] * dataL[offsetL + x + 1] + mix.CoeffsL[5] * dataL[offsetL + x + 0] + mix.CoeffsR[3] * dataR[offsetR + x + 2] + mix.CoeffsR[4] * dataR[offsetR + x + 1] + mix.CoeffsR[5] * dataR[offsetR + x + 0]);
						dataO[offsetO + x + 0] = (byte)(mix.CoeffsL[6] * dataL[offsetL + x + 2] + mix.CoeffsL[7] * dataL[offsetL + x + 1] + mix.CoeffsL[8] * dataL[offsetL + x + 0] + mix.CoeffsR[6] * dataR[offsetR + x + 2] + mix.CoeffsR[7] * dataR[offsetR + x + 1] + mix.CoeffsR[8] * dataR[offsetR + x + 0]);
					}
				}
			}

			return outputFrame;
		}

		/// <summary>Computes an anaglyph images from BGR32 images.</summary>
		/// <param name="bitmapL">Left input frame.</param>
		/// <param name="bitmapR">Right input frame.</param>
		/// <param name="mix">The colour mixing for the selected anaglyph method.</param>
		/// <returns>The anaglyph image, packaged as a BitmapFrame.</returns>
		private unsafe BitmapFrame ComputeAnaglyphBgr32(BitmapFrame bitmapL, BitmapFrame bitmapR, ColourMixing mix)
		{
			BitmapFrame outputFrame = new BitmapFrame(bitmapL.Width, bitmapL.Stride, bitmapL.Height, bitmapL.Format) { SequenceNumber = bitmapL.SequenceNumber };

			fixed (byte* dataL = bitmapL.Data, dataR = bitmapR.Data, dataO = outputFrame.Data)
			{
				for (int y = 0; y < outputFrame.Height; y++)
				{
					int offsetL = y * bitmapL.Stride;
					int offsetR = y * bitmapR.Stride;
					int offsetO = y * outputFrame.Stride;

					for (int x = 0; x < 4 * outputFrame.Width; x += 4)
					{
						// reminder: ARGB in uint32 is packed like 0123=BGRA
						dataO[offsetO + x + 2] = (byte)(mix.CoeffsL[0] * dataL[offsetL + x + 2] + mix.CoeffsL[1] * dataL[offsetL + x + 1] + mix.CoeffsL[2] * dataL[offsetL + x + 0] + mix.CoeffsR[0] * dataR[offsetR + x + 2] + mix.CoeffsR[1] * dataR[offsetR + x + 1] + mix.CoeffsR[2] * dataR[offsetR + x + 0]);
						dataO[offsetO + x + 1] = (byte)(mix.CoeffsL[3] * dataL[offsetL + x + 2] + mix.CoeffsL[4] * dataL[offsetL + x + 1] + mix.CoeffsL[5] * dataL[offsetL + x + 0] + mix.CoeffsR[3] * dataR[offsetR + x + 2] + mix.CoeffsR[4] * dataR[offsetR + x + 1] + mix.CoeffsR[5] * dataR[offsetR + x + 0]);
						dataO[offsetO + x + 0] = (byte)(mix.CoeffsL[6] * dataL[offsetL + x + 2] + mix.CoeffsL[7] * dataL[offsetL + x + 1] + mix.CoeffsL[8] * dataL[offsetL + x + 0] + mix.CoeffsR[6] * dataR[offsetR + x + 2] + mix.CoeffsR[7] * dataR[offsetR + x + 1] + mix.CoeffsR[8] * dataR[offsetR + x + 0]);
					}
				}
			}

			return outputFrame;
		}

		#endregion
	}
}
