﻿// $Id: DiffFilter.cs 278 2010-08-02 12:08:33Z cr333 $
using System;
using System.Collections.Generic;
using VideoLib.Frames;

namespace VideoLib.Filters
{
	/// <summary>Computes the difference between two bitmaps. Output format is the same like the "in0" input format.</summary>
	public class DiffFilter : Node
	{
		/// <summary>The value a difference of zero is mapped to. Default: 0.5.</summary>
		public double Offset { get; set; }

		/// <summary>Scaling factor to emphasise smaller differences. Default: 1.</summary>
		public double Scale { get; set; }

		/// <summary>Computes the difference image between [in0] and [in1].</summary>
		public DiffFilter()
			: base()
		{
			AddInputPins("in0", "in1");
			AddOutputPin("out");

			// default settings
			Offset = 0.5;
			Scale = 1;
		}

		/// <summary>Processes a video frame.</summary>
		public override void Process()
		{
			// get input frames
			var frames = GetInputFrames("in0", "in1");
			if (PassOnEndOfStreamFrame(frames)) return;
			var bitmaps = CastFramesTo<BitmapFrame>(frames);

			// compute difference image
			BitmapFrame outputFrame = ComputeDiff(bitmaps["in0"], bitmaps["in1"], (float)Offset, (float)Scale);

			// send off output frame
			PushFrame(outputFrame);
		}

		/// <summary>Computes the difference image between two bitmap frames (with the same layout of colour components).</summary>
		/// <param name="inputFrame1">The first input frame.</param>
		/// <param name="inputFrame2">The first input frame.</param>
		/// <param name="offset">The offset into the unit interval that a difference of zero is mapped to, for example 0.5.</param>
		/// <param name="scale">The scale factor applied to the difference. Often, this is 1.0.</param>
		public static BitmapFrame ComputeDiff(BitmapFrame inputFrame1, BitmapFrame inputFrame2, float offset, float scale)
		{
			// first make sure the frame have the same size and and colours
			AssertFrameProperties(new Dictionary<string, BitmapFrame>() { { "in0", inputFrame1 }, { "in1", inputFrame2 } }, FrameProps.Size | FrameProps.Layout);

			// get scaling factors
			int componentsPerRow = inputFrame1.Width * inputFrame1.Format.ComponentsPerPixel();
			var inputType1 = inputFrame1.Format & PixelFormat.TypeMask;
			var inputType2 = inputFrame2.Format & PixelFormat.TypeMask;
			float scaleIn1 = (float)(1.0 / inputType1.MaxValue());
			float scaleIn2 = (float)(1.0 / inputType2.MaxValue());
			float scaleOut = (float)(inputType1.MaxValue());

			// prepare output frame
			BitmapFrame outputFrame = new BitmapFrame(inputFrame1.Width, inputFrame1.Height, inputFrame1.Format);
			outputFrame.SequenceNumber = inputFrame1.SequenceNumber;

			// pin the input & output pointers
			unsafe
			{
				fixed (byte* pIn1 = inputFrame1.Data, pIn2 = inputFrame2.Data, pOut = outputFrame.Data)
				{
					for (int y = 0; y < inputFrame1.Height; y++)
					{
						// compute row pointers for both images (in bytes)
						byte* rIB1 = pIn1 + y * inputFrame1.Stride;
						byte* rIB2 = pIn2 + y * inputFrame2.Stride;
						byte* rOB  = pOut + y * outputFrame.Stride;

						// convert pointers to different datatypes
						ushort* rIS1 = (ushort*)rIB1;
						uint*   rII1 = (uint  *)rIB1;
						float*  rIF1 = (float *)rIB1;
						ushort* rIS2 = (ushort*)rIB2;
						uint*   rII2 = (uint  *)rIB2;
						float*  rIF2 = (float *)rIB2;
						ushort* rOS  = (ushort*)rOB;
						uint*   rOI  = (uint  *)rOB;
						float*  rOF  = (float *)rOB;

						// 1. read both input values
						// 2. scale them each to the unit interval
						// 3. apply global scale and offset
						// 4. scale and cast to output type
						if      (inputType1 == PixelFormat.Byte   && inputType2 == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte  )(scaleOut * (offset + scale * (scaleIn1 * (float)*rIB1++ - scaleIn2 * (float)*rIB2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Byte   && inputType2 == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte  )(scaleOut * (offset + scale * (scaleIn1 * (float)*rIB1++ - scaleIn2 * (float)*rIS2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Byte   && inputType2 == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte  )(scaleOut * (offset + scale * (scaleIn1 * (float)*rIB1++ - scaleIn2 * (float)*rII2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Byte   && inputType2 == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte  )(scaleOut * (offset + scale * (scaleIn1 * (float)*rIB1++ - scaleIn2 * (float)*rIF2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Short  && inputType2 == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(scaleOut * (offset + scale * (scaleIn1 * (float)*rIS1++ - scaleIn2 * (float)*rIB2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Short  && inputType2 == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(scaleOut * (offset + scale * (scaleIn1 * (float)*rIS1++ - scaleIn2 * (float)*rIS2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Short  && inputType2 == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(scaleOut * (offset + scale * (scaleIn1 * (float)*rIS1++ - scaleIn2 * (float)*rII2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Short  && inputType2 == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(scaleOut * (offset + scale * (scaleIn1 * (float)*rIS1++ - scaleIn2 * (float)*rIF2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Int    && inputType2 == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint  )(scaleOut * (offset + scale * (scaleIn1 * (float)*rII1++ - scaleIn2 * (float)*rIB2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Int    && inputType2 == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint  )(scaleOut * (offset + scale * (scaleIn1 * (float)*rII1++ - scaleIn2 * (float)*rIS2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Int    && inputType2 == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint  )(scaleOut * (offset + scale * (scaleIn1 * (float)*rII1++ - scaleIn2 * (float)*rII2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Int    && inputType2 == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint  )(scaleOut * (offset + scale * (scaleIn1 * (float)*rII1++ - scaleIn2 * (float)*rIF2++)) + 0.5f); } // round
						else if (inputType1 == PixelFormat.Single && inputType2 == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = (float )(scaleOut * (offset + scale * (scaleIn1 * (float)*rIF1++ - scaleIn2 * (float)*rIB2++))       ); }
						else if (inputType1 == PixelFormat.Single && inputType2 == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = (float )(scaleOut * (offset + scale * (scaleIn1 * (float)*rIF1++ - scaleIn2 * (float)*rIS2++))       ); }
						else if (inputType1 == PixelFormat.Single && inputType2 == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = (float )(scaleOut * (offset + scale * (scaleIn1 * (float)*rIF1++ - scaleIn2 * (float)*rII2++))       ); }
						else if (inputType1 == PixelFormat.Single && inputType2 == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = (float )(scaleOut * (offset + scale * (scaleIn1 * (float)*rIF1++ - scaleIn2 * (float)*rIF2++))       ); }
						else throw new NotImplementedException(String.Format("Conversion from \"{0}\" and \"{1}\" to \"{2}\" not implemented.", inputType1, inputType2, outputFrame.Format));
					}
				}
			}

			return outputFrame;
		}
	}
}
