﻿// $Id: PartitionSolver.cs 436 2010-10-15 16:26:38Z cr333 $
// Based on a Part II project by Rubin Xu (2010).
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace VideoLib.Scheduling
{
	/// <summary>
	/// A basic class for implementations of the number partitioning problem in the 
	/// context of job scheduling.
	/// </summary>
	public abstract class AbstractPartitionSolver
	{
		/// <summary>
		/// Decides if the input dataset is feasible to solve using the implemented algorithm.
		/// </summary>
		/// <param name="NumPartitions">Number of partitions</param>
		/// <param name="NumItems">Number of objects to partition</param>
		/// <returns></returns>
		public abstract bool IsProblemSizeReasonable(int NumPartitions, int NumItems);

		/// <summary>
		/// Solve the partitioning problem, return an optimal solution in the context of 
		/// the static job scheduling/assignment
		/// </summary>
		/// <param name="NumPartitions">Number of partitions c.f. available CPUs</param>
		/// <param name="Costs">Cost matrix for jobs</param>
		/// <param name="FixedItems">Artificial constraints on some of the jobs (GPU jobs)</param>
		/// <returns></returns>
		public abstract Dictionary<Object, int> FindOptimalAffinity(int NumPartitions, Dictionary<Object, double> Costs, HashSet<Object> FixedItems);
	}

#region Brute force Solver
	/// <summary>
	/// An unoptimized brute force searching algorithm. Currently not used anywhere.
	/// </summary>
	public class BruteForcePartitionSolver : AbstractPartitionSolver
	{
		/// <summary>
		/// Decides if the input dataset is feasible to solve using the implemented algorithm.
		/// </summary>
		/// <param name="NumPartitions">Number of partitions</param>
		/// <param name="NumItems">Number of objects to partition</param>
		/// <returns></returns>
		public override bool IsProblemSizeReasonable(int NumPartitions, int NumItems)
		{
			//The complexity is NumPartitions^(NumItems), and accept if this value is less than 2*10^7
			return Math.Pow((double)NumPartitions, (double)(NumItems)) <2e7;
		}

		/// <summary>
		/// Find an optimal assignment of affinity using brute force.
		/// The valid affinity range is 0.. AffinityRange
		/// </summary>
		/// <returns></returns>
		public override Dictionary<Object, int> FindOptimalAffinity(int NumPartitions, Dictionary<Object, double> Costs, HashSet<Object> FixedItems)
		{
			Dictionary<Object, int> BestAssignment = new Dictionary<Object, int>();
			LinkedList<Object> Nodes = new LinkedList<Object>();
			foreach (Object n in Costs.Keys)
				Nodes.AddFirst(n);
			double t= EnumerateAffinityAssignment(NumPartitions - 1, Costs, FixedItems, Nodes, new Dictionary<Object, int>(), BestAssignment, 1000000);
			return BestAssignment;
		}

		private double EnumerateAffinityAssignment(int AffinityRange, Dictionary<Object, double> Costs, HashSet<Object> FixedItems, 
													LinkedList<Object> UnassignedNodes, Dictionary<Object, int> Assignment,
													Dictionary<Object, int> BestAssignment, double BestTime)
		{
			// Still a partial assignment, recurse to complete it
			if (UnassignedNodes.Count != 0)
			{
				Object node = UnassignedNodes.First.Value;
				if (FixedItems.Contains(node)) //Fixed items can only have affinity 0
				{
					Assignment[node] = 0;
					UnassignedNodes.RemoveFirst();
					double t = EnumerateAffinityAssignment(AffinityRange, Costs, FixedItems, UnassignedNodes, Assignment,
														BestAssignment, BestTime);
					UnassignedNodes.AddFirst(node);
					return t;
				}
				else //Normal Object, enumerate all possible affinities
				{
					double t;
					for (int i = 0; i <= AffinityRange; i++)
					{
						Assignment[node] = i;
						UnassignedNodes.RemoveFirst();
						t = EnumerateAffinityAssignment(AffinityRange, Costs, FixedItems, UnassignedNodes, Assignment, BestAssignment, BestTime);
						UnassignedNodes.AddFirst(node);
						if (t < BestTime)
							BestTime = t;
					}
					return BestTime;
				}
			}
			else //Reaches a complete assignment so compute its score(Expected running time) and compare it with the current best known results.
			{
				double time = CalculateScore(Assignment, Costs);
				if (time < BestTime)
				{
					BestTime = time;
					BestAssignment.Clear();
					foreach (KeyValuePair<Object, int> pr in Assignment)
						BestAssignment.Add(pr.Key, pr.Value);
				}
				return BestTime;
			}
		}
		/// <summary>
		/// Given an Affinity assignment and computation costs for every node, output the expected total running time.
		/// </summary>
		/// <param name="Affinity"></param>
		/// <param name="Costs"></param>
		/// <returns></returns>
		private double CalculateScore(Dictionary<Object, int> Affinity, Dictionary<Object, double> Costs)
		{
			//Find total computation time for each thread assuming each thread goes round robin among its nodes.
			Dictionary<int, double> ThreadCosts = new Dictionary<int, double>();
			foreach (KeyValuePair<Object, int> pr in Affinity)
			{
				if (ThreadCosts.ContainsKey(pr.Value))
					ThreadCosts[pr.Value] += Costs[pr.Key];
				else
					ThreadCosts[pr.Value] = Costs[pr.Key];
			}

			//The maximum time among all the ThreadCosts will become the expected time between consecutive frames produced.
			return ThreadCosts.Values.Max();
		}

	}
#endregion

	/// <summary>
	/// A heuristic search named complete greedy algorithm. 
	/// </summary>
	public class CompleteGreedyAlgorithmPartitionSolver : AbstractPartitionSolver
	{
		/// <summary>
		/// In case of big dataset, could set this value to speed up searching time
		/// at the cost of sub-optimal solution. 
		/// (SubOptimal - Optimal) will be smaller than  AllowedErrorRange 
		/// </summary>
		public double AllowedErrorRange { get; set; }

		/// <summary>
		/// Initializes a new instance of the <see cref="CompleteGreedyAlgorithmPartitionSolver"/> class.
		/// </summary>
		public CompleteGreedyAlgorithmPartitionSolver() 
		{
			// 0.1 ms error margin allowed
			AllowedErrorRange = 0.0001;
		}

		/// <summary>
		/// accept all datasets
		/// </summary>
		/// <param name="NumPartitions"></param>
		/// <param name="NumItems"></param>
		/// <returns></returns>
		public override bool IsProblemSizeReasonable(int NumPartitions, int NumItems)
		{
			// Currently this is the only remaining option if brute force is not feasible.
			return true;
		}

		/// <summary>
		/// Find optimal affinity
		/// </summary>
		/// <param name="NumPartitions"></param>
		/// <param name="Costs"></param>
		/// <param name="FixedItems"></param>
		/// <returns></returns>
		public override Dictionary<Object, int> FindOptimalAffinity(int NumPartitions, Dictionary<Object, double> Costs, HashSet<Object> FixedItems)
		{
			//First create a map between objects to int 0..NumOfObjects-1 to make life easier.
			//Since CGA processes items from largest weight to smallest weight, here we first sort
			//all objects according to their weights and assign indices sequentially from largest weight
			//to smallest weight.
			//However, those objects in FixedItems are assigned first regardless of their weights, this is 
			//to make CGASearch look at them as soon as possible to reduce the search tree's size.
			Dictionary<Object, int> ObjectMapping = new Dictionary<Object, int>();
			int NumObjects = 0;
			IEnumerable<Object> objects = Costs.Keys.Except(FixedItems);
			foreach(Object obj in FixedItems)
				ObjectMapping[obj] = NumObjects++;
			foreach(Object obj in objects.OrderByDescending(x => Costs[x]))
				ObjectMapping[obj] = NumObjects++;

			
			//Now convert all input object collections to arrays.
			double[] CostMatrix = new double[NumObjects];
			Boolean[] IsFixed = new Boolean[NumObjects];
			double Sum = 0.0;

			foreach (KeyValuePair<Object, double> pr in Costs)
			{
				CostMatrix[ObjectMapping[pr.Key]] = pr.Value;
				Sum += pr.Value;
				IsFixed[ObjectMapping[pr.Key]] = FixedItems.Contains(pr.Key);
			}

			this.NumObjects = NumObjects;
			this.NumPartitions = NumPartitions;
			this.Costs = CostMatrix;
			this.IsFixed = IsFixed;
			this.SumAllNumbers = Sum;
			double[] CPW = new double[NumPartitions];
			int[] BestPartition = new int[NumObjects];
			double best = CGASearch(0, new int[NumObjects], CPW, BestPartition, Sum);

			Dictionary<Object, int> par = new Dictionary<Object, int>();
			foreach (KeyValuePair<Object, int> pr in ObjectMapping)
				par[pr.Key] = BestPartition[pr.Value];
			return par;
		}

		//Global values used by GPASearch. Due to these variables, this class is not thread-safe
		private int NumObjects;
		private int NumPartitions;
		private double[]Costs;
		private Boolean[] IsFixed;
		private double SumAllNumbers;
		private double[] Test1 = new double[100];

		// The return value is the difference between the maximum weighted partition and the minimum weighted
		// partition
		private double CGASearch(int CurLevel, int[] CurrentPartition, double[] CurrentPartitionWeights, 
								 int[] BestPartition, double BestScore)
		{
				if (CurLevel <= NumObjects - 1) //Still in the branch of the tree
				{
					//At current level curLevel, the tree is expanded with NumPartitions branches
					//The ith branch represents that the Object indexed by curLevel is allocated to partition i

					if (IsFixed[CurLevel]) //Fixed object, allocate it to Partition 0
					{
						CurrentPartition[CurLevel] = 0;
						CurrentPartitionWeights[0] += Costs[CurLevel];
						return CGASearch(CurLevel + 1, CurrentPartition, CurrentPartitionWeights, BestPartition, BestScore);

					}

					// Pruning rule
					double CurLargestSum = -1;
					for (int i = 0; i < NumPartitions; i++)
						if (CurrentPartitionWeights[i] > CurLargestSum) CurLargestSum = CurrentPartitionWeights[i];
					if (CurLargestSum - ((SumAllNumbers - CurLargestSum) / (NumPartitions - 1)) > BestScore)
						return BestScore;

					// Allocate Obj[CurLevel] to partitions in order of ascending CurrentPartitionWeights
					int[] Order = new int[NumPartitions];
					for (int i = 0; i < NumPartitions; i++) Order[i] = i;
					for (int i = 0; i < NumPartitions; i++) //Uses a quadratic sorting algorithm..sufficient here as NumPartitions<=4
						for (int j = i + 1; j < NumPartitions; j++)
							if (CurrentPartitionWeights[Order[i]] > CurrentPartitionWeights[Order[j]])
							{
								int k = Order[i];
								Order[i] = Order[j];
								Order[j] = k;
							}
					double BestS = BestScore;
					for (int i = 0; i < NumPartitions; i++) //Allocate Obj[CurLevel] to partition Order[i]
					{
						int p = Order[i];
						CurrentPartition[CurLevel] = p;
						double t = CurrentPartitionWeights[p];
						CurrentPartitionWeights[p] += Costs[CurLevel];
						double NewBestS = CGASearch(CurLevel + 1, CurrentPartition, CurrentPartitionWeights, BestPartition, BestS);
						if (NewBestS < BestS) BestS = NewBestS;
						CurrentPartitionWeights[p] = t;

						if (NewBestS < AllowedErrorRange)  // Pruning.
							return NewBestS;
					}
					return BestS;
				}
				else //Reach a leaf of the tree.
				{
					double min = 1000000.0;
					double max = 0;
					for (int i = 0; i < NumPartitions; i++)
					{
						if (CurrentPartitionWeights[i] > max) max = CurrentPartitionWeights[i];
						if (CurrentPartitionWeights[i] < min) min = CurrentPartitionWeights[i];
					}

					if (max - min < BestScore) //Found a better solution
					{
						for (int i = 0; i < NumObjects; i++)
							BestPartition[i] = CurrentPartition[i];
						for (int i = 0; i < NumPartitions; i++)
							Test1[i] = CurrentPartitionWeights[i];
						return max - min;
					}
					else
					{
						return BestScore;
					}
				}

		}
	
	}

}
