﻿// $Id: Profiler.cs 293 2010-08-04 16:26:33Z cr333 $
// Based on a Part II project by Rubin Xu (2010).
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Diagnostics;

namespace VideoLib.Scheduling
{
	/// <summary>
	/// Represents a collection of profiling data on a given target node/object
	/// </summary>
	public class ProfilingData
	{
		/// <summary>The profiling target.</summary>
		public Object Target { get; protected set; }

		/// <summary>Mean execution time.</summary>
		public double RunTimeMean { get; protected set; }

		/// <summary>Standard deviation of execution times.</summary>
		public double RunTimeStDev { get; protected set; }

		/// <summary>The collection of execution times (in seconds).</summary>
		public IEnumerable<double> RunTimes { get; protected set; }

		/// <summary>The collection of times between consecutive execution starts. Indicator of framerate.</summary>
		public IEnumerable<double> Intervals { get; protected set; }

		private IEnumerable<long> rawdata;

		/// <summary>Writes the profiling data to disk.</summary>
		/// <param name="prefix">Filename prefix.</param>
		public void WriteDataToDisk(string prefix)
		{
			StreamWriter logfile = File.CreateText(prefix+Target.ToString()+".txt");

			foreach(long d in rawdata)
				logfile.WriteLine(String.Format("{0}", d));

			logfile.Close();
		}

		/// <summary>Loads the profiling data from disk.</summary>
		/// <param name="logfilename">Filename of the log file.</param>
		public void LoadDataFromDisk(string logfilename)
		{
			StreamReader logfile = File.OpenText(logfilename);
			//Create a fake String target.
			Target = Path.GetFileNameWithoutExtension(logfilename);

			LinkedList<long> timepoints  = new LinkedList<long>();
			while (!logfile.EndOfStream)
			{
				timepoints.AddLast(long.Parse(logfile.ReadLine()));
			}
			logfile.Close();

			ProcessProfilingResults(timepoints);
		}

		/// <summary>Calculates the profiling data given a list of 2n time points, where each pair
		/// (timepoints[2i], timepoints[2i+1]) represents an interval during which the target was running.</summary>
		public void ProcessProfilingResults(LinkedList<long> timepoints)
		{
			if (timepoints.Count < 2) return;

			rawdata = new List<long>(timepoints);
			long lastt1 = -1; // remember previous start time (for FPS)

			LinkedList<double> runtimes = new LinkedList<double>();
			LinkedList<double> intervals = new LinkedList<double>();

			while (timepoints.Count >= 2)
			{
				long t1 = timepoints.First.Value; timepoints.RemoveFirst(); // start time
				long t2 = timepoints.First.Value; timepoints.RemoveFirst(); // end time

				// time between MarkEnd and MarkBegin: execution time measurement
				runtimes.AddLast((double)(t2 - t1) / Stopwatch.Frequency);

				// time between consecutive MarkBegins: framerate measurement
				if (lastt1 >= 0) intervals.AddLast((double)(t1 - lastt1) / Stopwatch.Frequency);

				lastt1 = t1;
			}

			RunTimeMean  = runtimes.Average();
			RunTimeStDev = Math.Sqrt(runtimes.Average(x => (x - RunTimeMean) * (x - RunTimeMean)));
			RunTimes  = runtimes;
			Intervals = intervals;
		}

		/// <summary>Initializes a new instance of the <see cref="ProfilingData"/> class.</summary>
		/// <param name="target">The profiling target.</param>
		/// <param name="timepoints">The timepoints.</param>
		public ProfilingData(Object target, LinkedList<long> timepoints)
		{
			Target = target;
			ProcessProfilingResults(timepoints);
		}

		/// <summary>Create an instance from a log file (possibly created by WriteDataToDisk previously).</summary>
		/// <param name="StoredLogFile"></param>
		public ProfilingData(String StoredLogFile)
		{
			Target = null;
			LoadDataFromDisk(StoredLogFile);
		}
	}

	/// <summary>Profiler interface.</summary>
	public interface IProfiler
	{
		/// <summary>Resets this profiling data.</summary>
		void Reset();

		/// <summary>Marks the begin of a target object's execution.</summary>
		/// <param name="target">The target object.</param>
		void MarkStarting(Object target);

		/// <summary>Marks the finish of a target object's execution</summary>
		/// <param name="target">The target object.</param>
		void MarkEnding(Object target);

		/// <summary>Prepares the results.</summary>
		void PrepareResults();

		/// <summary>Returns the profiling data of a target object.</summary>
		/// <param name="target">The target object.</param>
		/// <returns>Profiling data for the given target object.</returns>
		ProfilingData GetResult(Object target);

		/// <summary>Returns all registered profiling targets.</summary>
		/// <returns>Set of profiled objects.</returns>
		IEnumerable<Object> GetProfilingTargets();
	}

	/// <summary>
	/// One concrete implementation for profiler
	/// </summary>
	public class BasicProfiler : IProfiler
	{
		private LinkedList<KeyValuePair<Object, long>> times = new LinkedList<KeyValuePair<Object, long>>();

		private Dictionary<Object, ProfilingData> results = new Dictionary<Object, ProfilingData>();

		/// <summary>
		/// Clear all stored profiling data
		/// </summary>
		public void Reset()
		{
			lock(times){
				times.Clear();
			}
		}

		/// <summary>
		/// Called to mark the beginning of target's computation period
		/// </summary>
		/// <param name="target"></param>
		public void MarkStarting(Object target)
		{
			long time = Stopwatch.GetTimestamp();
			lock (times)
			{
				times.AddLast(new KeyValuePair<Object, long>(target, time));
			}
		}

		/// <summary>
		/// Called to mark the end of target's computation period
		/// </summary>
		/// <param name="target"></param>
		public void MarkEnding(Object target)
		{
			long time = Stopwatch.GetTimestamp();
			lock (times)
			{
				times.AddLast(new KeyValuePair<Object, long>(target, time));
			}
		}

		/// <summary>
		/// Compute profiling results based on all collected values 
		/// </summary>
		public void PrepareResults()
		{
			Dictionary<Object, LinkedList<long>> data = new Dictionary<Object, LinkedList<long>>();
			lock(times)
			{
				results.Clear();
				foreach (KeyValuePair<Object, long> timepoint in times)
				{
					if (!data.ContainsKey(timepoint.Key))
						data[timepoint.Key] = new LinkedList<long>();
					data[timepoint.Key].AddLast(timepoint.Value);
				}

				lock(results)
				{
					results.Clear();
					foreach (KeyValuePair<Object, LinkedList<long>> target in data)
					{
						results[target.Key] = new ProfilingData(target.Key, target.Value);
					}
				}
			}
		}

		/// <summary>
		///  return profiling results for a given target
		/// </summary>
		public ProfilingData GetResult(Object target)
		{
			return results[target];
		}

		/// <summary>
		/// return a collection of objects that hav profiling data collected.
		/// </summary>
		/// <returns></returns>
		public IEnumerable<Object> GetProfilingTargets()
		{
			return results.Keys;
		}
	}
}
