﻿// $Id: ImageSequenceSink.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing.Imaging;
using VideoLib.Frames;
using System.IO;

namespace VideoLib.Sinks
{
	/// <summary>Writes a sequence of images to disk.</summary>
	public class ImageSequenceSink : Node
	{
		/// <summary>Filename template for use with String.Format, e.g. "{0:D4}.png" for "NNNN.png".</summary>
		public string Filenames { get; protected set; }

		/// <summary>Image format used for storing the images.</summary>
		public ImageFormat ImageFormat { get; protected set; }

		/// <summary>Human-friendly node name.</summary>
		public override string Name
		{
			get { return String.Format("{0}({1})", base.Name, Path.GetFileName(String.Format(Filenames, 1234))); }
		}

		/// <summary>Constructs a new image sequence sink.</summary>
		/// <param name="filenames">Template for filenames for use with String.Format, e.g. "{0:D4}.png" for "NNNN.png".</param>
		public ImageSequenceSink(string filenames) : base()
		{
			Filenames = filenames;
			AddInputPin("in");

			// guess image format from file extension
			ImageFormat = GuessFormat();
		}

		/// <summary>Processes a frame.</summary>
		public override void Process()
		{
			// get input frame
			Frame inputFrame = GetInputFrame("in");

			// ignore EOS frames
			if (inputFrame is EndOfStreamFrame)
			{
				Status = NodeStatus.Finished;
				return;
			}

			// make sure directory exists
			string filename = String.Format(Filenames, inputFrame.SequenceNumber);
			if (!madeDirs) MakeDirs(filename);

			// save image
			BitmapFrame bitmapFrame = CastFrameTo<BitmapFrame>(inputFrame);
			bitmapFrame.Bitmap.Save(filename, ImageFormat);
		}

		/// <summary>Stores if the directory, where the images are kept, has been created (if necessary).</summary>
		private bool madeDirs = false;

		/// <summary>Guesses the image format from the extension of the filename template.</summary>
		/// <returns>Image format.</returns>
		private ImageFormat GuessFormat()
		{
			string someFilename = String.Format(Filenames, 0);
			string extension = new FileInfo(someFilename).Extension.ToLowerInvariant();

			switch (extension)
			{
				case ".bmp":  return ImageFormat.Bmp;
				case ".gif":  return ImageFormat.Gif;
				case ".jpg":  return ImageFormat.Jpeg;
				case ".jpeg": return ImageFormat.Jpeg;
				case ".png":  return ImageFormat.Png;
				case ".tif":  return ImageFormat.Tiff;
				case ".tiff": return ImageFormat.Tiff;
				default:      return ImageFormat.Png;
			}
		}

		/// <summary>Makes sure that the image directory exists.</summary>
		/// <param name="filename">One image path.</param>
		private void MakeDirs(string filename)
		{
			FileInfo fileInfo = new FileInfo(filename);
			DirectoryInfo parentDir = fileInfo.Directory;
			if (!parentDir.Exists)
				parentDir.Create();
			madeDirs = true;
		}
	}
}
