﻿// $Id: RawSink.cs 552 2012-04-04 14:07:06Z cr333 $
using System;
using System.IO;
using VideoLib.Frames;
using System.Collections.Generic;

namespace VideoLib.Sinks
{
	/// <summary>Writes a sequence of raw frames to disk.</summary>
	public class RawSink : Node
	{
		/// <summary>Filename template for use with String.Format, e.g. "{0:D4}.png" for "NNNN.png".</summary>
		public string Filenames { get; protected set; }

		/// <summary>Gets or sets whether writing to disk should be asynchronous.</summary>
		public bool Async { get; set; }

		/// <summary>Gets or sets whether frames are buffered in memory and only written disk when the scheduler stops.</summary>
		public bool BufferFrames { get; set; }

		/// <summary>Collection of all frames buffered in memory.</summary>
		private Dictionary<string, BitmapFrame> frameStore;

		/// <summary>Human-friendly node name.</summary>
		public override string Name
		{
			get { return String.Format("{0}({1})", base.Name, Path.GetFileName(String.Format(Filenames, 1234))); }
		}

		/// <summary>Constructs a new raw frame sink.</summary>
		/// <param name="filenames">Template for filenames for use with String.Format, e.g. "{0:D4}.png" for "NNNN.png".</param>
		public RawSink(string filenames)
			: base()
		{
			Async        = false;
			BufferFrames = false;
			Filenames    = filenames;

			AddInputPin("in");
		}

		/// <summary>Initialisation code, executed just before the first StatusUpdate.</summary>
		public override void Init()
		{
			// make sure directory exists
			MakeDirs(String.Format(Filenames, 0));

			// if buffering in memory: create frame store
			if(BufferFrames)
				frameStore = new Dictionary<string, BitmapFrame>();
		}

		/// <summary>Processes a frame.</summary>
		public override void Process()
		{
			// get input frame
			Frame inputFrame = GetInputFrame("in");

			// handle EOS frame
			if (inputFrame is EndOfStreamFrame)
			{
				Status = NodeStatus.Finished;
				return;
			}

			// save frame
			BitmapFrame bitmapFrame = CastFrameTo<BitmapFrame>(inputFrame);
			string filename = String.Format(Filenames, bitmapFrame.SequenceNumber);
			SaveFrame(bitmapFrame, filename);
		}

		/// <summary>Cleanup code, executed when stopping the scheduler.</summary>
		public override void Stop()
		{
			if (BufferFrames)
			{
				// make sure directory exists
				MakeDirs(String.Format(Filenames, 0));

				// write buffered frames to disk
				foreach (var frame in frameStore)
					SaveFrameToFile(frame.Value, frame.Key);

				// some cleanup
				frameStore.Clear();
				frameStore = null;
			}

			base.Stop();
		}

		/// <summary>Makes sure that the image directory exists.</summary>
		/// <param name="filename">One image path.</param>
		private void MakeDirs(string filename)
		{
			FileInfo fileInfo = new FileInfo(filename);
			DirectoryInfo parentDir = fileInfo.Directory;
			if (!parentDir.Exists)
				parentDir.Create();
		}

		/// <summary>Saves the given frame to memory or disk (depending on settings).</summary>
		/// <param name="bitmapFrame">The bitmap frame.</param>
		/// <param name="filename">The filename.</param>
		private void SaveFrame(BitmapFrame bitmapFrame, string filename)
		{
			if (BufferFrames)
			{
				frameStore.Add(filename, bitmapFrame);
			}
			else
			{
				SaveFrameToFile(bitmapFrame, filename);
			}
		}

		
		/// <summary>Saves the given frame to a file.</summary>
		/// <param name="bitmapFrame">The bitmap frame.</param>
		/// <param name="filename">The filename.</param>
		private void SaveFrameToFile(BitmapFrame bitmapFrame, string filename)
		{
			SaveFrameToFile(bitmapFrame, filename, Async);
		}

		/// <summary>Saves the given frame to a file.</summary>
		/// <param name="bitmapFrame">The bitmap frame.</param>
		/// <param name="filename">The filename.</param>
		/// <param name="async">Use asynchronous IO if true.</param>
		public static void SaveFrameToFile(BitmapFrame bitmapFrame, string filename, bool async)
		{
			// Note that the final parameter to FileStream says that the write should be synchronous,
			// but it will still be executed asynchronously in a different thread.
			// This performed better than using asynchronous file IO.
			var stream = new FileStream(filename, FileMode.Create, FileAccess.Write, FileShare.Write, 1<<13, false);
			var writer = new BinaryWriter(stream);

			writer.Write("BitmapFrame");
			writer.Write((UInt32)1); // version of raw format used
			writer.Write((Int32)bitmapFrame.Width);
			writer.Write((Int32)bitmapFrame.Height);
			writer.Write((UInt32)bitmapFrame.Format);
			writer.Write((Int32)bitmapFrame.Stride);

			if (async)
			{
				// start writing asynchronously and call the callback when done
				stream.BeginWrite(bitmapFrame.Data, 0, bitmapFrame.Data.Length,
					new AsyncCallback(SaveToFileCallback), stream);

				// the stream is closed by callback function
			}
			else
			{
				writer.Write(bitmapFrame.Data);
				writer.Close();
				stream.Close();
			}
		}

		private static void SaveToFileCallback(IAsyncResult asyncResult)
		{
			var stream = (FileStream)asyncResult.AsyncState;
			stream.EndWrite(asyncResult);
			stream.Close();
		}
	}
}
