﻿// $Id: ImageSequenceSource.cs 531 2011-04-19 15:57:11Z ls426 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;
using System.IO;
using System.Drawing;
using System.Threading;

// 2010-07-27 CR -- Note on using timers
// -------------------------------------
// I tried timers (both System.Timers.Timer and System.Threading.Timer),
// but neither managed to get close to the desired framerate. For example,
// they only managed a sustained 64 fps when set to 10 ms interval (100 Hz).

namespace VideoLib.Sources
{
	/// <summary>Loads a sequence of images from disk.</summary>
	public class ImageSequenceSource : Node
	{
		/// <summary>Filename template for use with String.Format, e.g. "{0:D4}.png" for "NNNN.png".</summary>
		public string Filenames { get; protected set; }

		/// <summary>The sequence number of the most recently loaded image.</summary>
		public int SequenceNumber { get; protected set; }

		/// <summary>The framerate with which the video frames will be output.</summary>
		public float FPS { get; set; }

		/// <summary>Sequence number of the first frame. Default: 0.</summary>
		public int FirstNumber { get; set; }

		/// <summary>Sequence number of the last frame. Default: -1 (infinite).</summary>
		public int LastNumber { get; set; }

		/// <summary>Restart sequence after the last frame. Default: false.</summary>
		public bool Loop { get; set; }

		/// <summary>Human-friendly node name.</summary>
		public override string Name
		{
			get { return String.Format("{0}({1})", base.Name, Path.GetFileName(String.Format(Filenames, 1234))); }
		}

		/// <summary>Maximum number of missing images in the image sequence. Default: 0 (no missing images).</summary>
		public int MaxGap { get; set; }

		/// <summary>Last frame that was loaded successfully.</summary>
		private int lastFrameNumber;

		/// <summary>Time of the last frame, to control FPS.</summary>
		private DateTime lastFrameTime = DateTime.Now;

		/// <summary>Constructs a new image sequence source.</summary>
		/// <param name="filenames">Template for filenames for use with String.Format, e.g. "{0:D4}.png" for "NNNN.png".</param>
		public ImageSequenceSource(string filenames) : base()
		{
			// Default settings
			Filenames = filenames;
			FirstNumber = 0;
			LastNumber = -1;
			MaxGap = 0;
			FPS = 25;
			Loop = false;

			AddOutputPin("out");
		}

		/// <summary>Prepares the image sequence source.</summary>
		public override void Init()
		{
			SequenceNumber = FirstNumber;
		}

		private static object locker = new object();

		/// <summary>Processes a frame.</summary>
		public override void Process()
		{
			// read in the next image, skipping missing ones if required
			do
			{
				// check if we're still in the valid range of sequence numbers
				if (LastNumber != -1 && SequenceNumber > LastNumber)
				{
					if (Loop)
					{
						SequenceNumber = FirstNumber;
					}
					else
					{
						PushFrame("out", new EndOfStreamFrame() { SequenceNumber = SequenceNumber });
						Status = NodeStatus.Finished;
						return;
					}
				}

				// filename of the current image
				string filename = String.Format(Filenames, SequenceNumber);

				if (File.Exists(filename))
				{
					Bitmap bitmap;

					// Trying to load two bitmaps at the same time can cause access violations. This is because GDI+
					// needs to be initialised before it can be used, and while one thread is duly initialising GDI+,
					// the other one thinks that GDI+ has already been initialised. When it opens the Bitmap, it will
					// then throw an AccessViolationException. Surrounding the Bitmap constructor with a static lock
					// serialises access and thus works around this issue. For further detail, please refer to
					// <http://connect.microsoft.com/VisualStudio/feedback/details/237718/system-drawing-classes-initialize-gdi-in-a-non-thread-safe-manner>
					// TODO: turn this into a public static property, and only lock on first image load(?).
					lock (locker)
					{
						// load image and package as BitmapFrame
						bitmap = new Bitmap(filename);
					}

					BitmapFrame frame = new BitmapFrame(bitmap) { SequenceNumber = SequenceNumber };

					// push new frame, updating internal state
					PushFrame("out", frame);
					lastFrameNumber = SequenceNumber;
					SequenceNumber++;

					// sleep to get framerate right
					Thread.Sleep((int)Math.Max(0, 1000 / FPS - (DateTime.Now.Ticks - lastFrameTime.Ticks) / 1e4));
					lastFrameTime = DateTime.Now;

					return;
				}
				else
				{
					// skip missing image
					SequenceNumber++;
					//continue;
				}

				// check if the gap has been too long
				if (SequenceNumber > lastFrameNumber + MaxGap + 1)
				{
					if (Loop)
					{
						SequenceNumber = FirstNumber;
					}
					else
					{
						PushFrame("out", new EndOfStreamFrame() { SequenceNumber = SequenceNumber });
						Status = NodeStatus.Finished;
						return;
					}
				}
			} while (SequenceNumber <= lastFrameNumber + MaxGap + 1);

			//// if control reaches this, the gap has been too long, so we end here
			//PushFrame("out", new EndOfStreamFrame() { SequenceNumber = SequenceNumber });
			//Status = NodeStatus.Finished;
		}
	}
}
